@@ -13,6 +13,7 @@ from .utils.misc import (
     log_step,
     get_pyxis_ssl_paths,
 )
+from .quay_client import QuayClient
 
 LOG = logging.getLogger("pubtools.quay")
 
@@ -24,18 +25,21 @@ class OperatorPusher:
     No validation is performed, push items are expected to be correct.
     """
 
-    def __init__(self, push_items, target_settings):
+    def __init__(self, push_items, task_id, target_settings):
         """
         Initialize.
 
         Args:
             push_items ([ContainerPushItem]):
                 List of push items.
+            task_id (str):
+                task id
             target_settings (dict):
                 Target settings.
         """
         self.push_items = push_items
         self.target_settings = target_settings
+        self.task_id = task_id
 
         self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
         self._version_items_mapping = {}
@@ -152,12 +156,15 @@ class OperatorPusher:
         """
         Get bundles to be deprecated in the index image.
 
+        If deprecation list URL isn't in the target settings, None is returned.
+
         Args:
             version: (str)
                 version for which deprecation list will be fetched.
 
         Returns:
-            list(str): list of bundles to be deprecated in the index image.
+            list(str)|None: list of bundles to be deprecated in the index image. or None if
+                            deprecation list URL was not specified.
         """
 
         def _get_requests_session():
@@ -175,6 +182,9 @@ class OperatorPusher:
 
             return session
 
+        if not self.target_settings["iib_deprecation_list_url"]:
+            return None
+
         deprecation_list = []
         deprecation_list_url = "{0}/{1}.yml/raw?ref=master".format(
             self.target_settings["iib_deprecation_list_url"].rstrip("/"), version.replace(".", "_")
@@ -183,7 +193,7 @@ class OperatorPusher:
 
         LOG.info("Getting the deprecation list for OCP version {0}".format(version))
         session = _get_requests_session()
-        response = session.get(url=deprecation_list_url)
+        response = session.get(url=deprecation_list_url, timeout=10)
         if not response.ok:
             LOG.error(
                 "Could not retrieve deprecation list after multiple attempts."
@@ -224,7 +234,7 @@ class OperatorPusher:
         args += ["--iib-server", target_settings["iib_server"]]
         args += ["--iib-krb-principal", target_settings["iib_krb_principal"]]
 
-        if "iib_overwrite_from_index" in target_settings:
+        if target_settings.get("iib_overwrite_from_index", False):
             args += ["--overwrite-from-index"]
         if "iib_krb_ktfile" in target_settings:
             args += ["--iib-krb-ktfile", target_settings["iib_krb_ktfile"]]
@@ -239,7 +249,13 @@ class OperatorPusher:
 
     @classmethod
     def iib_add_bundles(
-        cls, bundles=None, archs=None, index_image=None, deprecation_list=None, target_settings={}
+        cls,
+        bundles=None,
+        archs=None,
+        index_image=None,
+        deprecation_list=None,
+        build_tags=None,
+        target_settings={},
     ):
         """
         Construct and execute pubtools-iib command to add bundles to index image.
@@ -253,6 +269,8 @@ class OperatorPusher:
                 Index image to add the bundles to.
             deprecation_list ([str]|str):
                 List of bundles to be deprecated. Accepts both str (csv) and a list.
+            build_tags ([str]):
+                Extra tags that the new index image should be tagged with.
             target_settings (dict):
                 Settings used for setting the value of pubtools-iib parameters.
 
@@ -277,16 +295,24 @@ class OperatorPusher:
             args += ["--deprecation-list", deprecation_list]
         elif deprecation_list and isinstance(deprecation_list, list):
             args += ["--deprecation-list", ",".join(deprecation_list)]
+        if build_tags:
+            for build_tag in build_tags:
+                args += ["--build-tag", build_tag]
 
-        return run_entrypoint(
-            ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
-            "pubtools-iib-add-bundles",
-            args,
-            env_vars,
-        )
+        try:
+            return run_entrypoint(
+                ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
+                "pubtools-iib-add-bundles",
+                args,
+                env_vars,
+            )
+        except SystemExit:
+            return False
 
     @classmethod
-    def iib_remove_operators(cls, operators=None, archs=None, index_image=None, target_settings={}):
+    def iib_remove_operators(
+        cls, operators=None, archs=None, index_image=None, build_tags=None, target_settings={}
+    ):
         """
         Construct and execute pubtools-iib command to remove operators from index image.
 
@@ -295,8 +321,10 @@ class OperatorPusher:
                 Operator names to be removed from the index image.
             archs ([str]):
                 Architectures to build for.
-            ocp_version (str):
+            index_image (str):
                 Index image to remove the operators from.
+            build_tags ([str]):
+                Extra tags that the new index image should be tagged with.
             target_settings (dict):
                 Settings used for setting the value of pubtools-iib parameters.
 
@@ -318,6 +346,9 @@ class OperatorPusher:
         if archs:
             for arch in archs:
                 args += ["--arch", arch]
+        if build_tags:
+            for build_tag in build_tags:
+                args += ["--build-tag", build_tag]
 
         return run_entrypoint(
             ("pubtools-iib", "console_scripts", "pubtools-iib-remove-operators"),
@@ -349,7 +380,9 @@ class OperatorPusher:
         for version in sorted(self.version_items_mapping.keys()):
             image_ref = "{0}:{1}".format(index_image_repo, version)
             try:
-                manifest_list = quay_client.get_manifest(image_ref, manifest_list=True)
+                manifest_list = quay_client.get_manifest(
+                    image_ref, media_type=QuayClient.MANIFEST_LIST_TYPE
+                )
             except requests.exceptions.HTTPError as e:
                 if e.response.status_code == 404 or e.response.status_code == 401:
                     continue
@@ -406,6 +439,7 @@ class OperatorPusher:
                 archs=archs,
                 index_image=index_image,
                 deprecation_list=deprecation_list,
+                build_tags=["{0}-{1}".format(index_image.split(":")[1], self.task_id)],
                 target_settings=self.target_settings,
             )
 
@@ -425,6 +459,7 @@ class OperatorPusher:
                 extra tag suffix applied to iib version tags if specified
         """
         image_schema = "{host}/{namespace}/{repo}"
+        image_schema_tag = "{host}/{namespace}/{repo}:{tag}"
         index_image_repo = image_schema.format(
             host=self.quay_host,
             namespace=self.target_settings["quay_namespace"],
@@ -432,15 +467,25 @@ class OperatorPusher:
         )
 
         for version in self.version_items_mapping:
-            build_details = iib_results[version]["iib_result"]
+            build_details = iib_results.get(version, {}).get("iib_result", None)
+            if not build_details:
+                continue
 
             _, tag = build_details.index_image.split(":", 1)
+            permanent_index_image = image_schema_tag.format(
+                host=build_details.index_image_resolved.split("/")[0],
+                namespace=build_details.index_image_resolved.split("/")[1],
+                repo="iib",
+                tag=build_details.build_tags[0],
+            )
             dest_image = "{0}:{1}".format(index_image_repo, tag)
+            # We don't use permanent index image here because we always want to overwrite
+            # production tags with the latest index image (in case of parallel pushes)
             ContainerImagePusher.run_tag_images(
                 build_details.index_image, [dest_image], True, self.target_settings
             )
             if tag_suffix:
                 dest_image = "{0}:{1}-{2}".format(index_image_repo, tag, tag_suffix)
                 ContainerImagePusher.run_tag_images(
-                    build_details.index_image, [dest_image], True, self.target_settings
+                    permanent_index_image, [dest_image], True, self.target_settings
                 )
