@@ -1,5 +1,7 @@
 from collections import namedtuple
 import logging
+import sys
+from time import sleep
 
 import requests
 
@@ -24,7 +26,7 @@ LOG = logging.getLogger("pubtools.quay")
 class PushDocker:
     """Handle full Docker push workflow."""
 
-    ImageData = namedtuple("ImageData", ["repo", "tag", "digest"])
+    ImageData = namedtuple("ImageData", ["repo", "tag", "digest", "v2s1_digest"])
 
     def __init__(self, push_items, hub, task_id, target_name, target_settings):
         """
@@ -105,19 +107,6 @@ class PushDocker:
                 raise InvalidTargetSettings(
                     "'{0}' must be present in the docker settings.".format(setting)
                 )
-        if (
-            "iib_overwrite_from_index_token" in self.target_settings
-            and "iib_overwrite_from_index" not in self.target_settings
-        ) or (
-            "iib_overwrite_from_index_token" not in self.target_settings
-            and "iib_overwrite_from_index" in self.target_settings
-        ):
-            msg = (
-                "Either both or neither of 'iib_overwrite_from_index' and "
-                "'iib_overwrite_from_index_token' should be specified in target settings."
-            )
-            LOG.error(msg)
-            raise InvalidTargetSettings(msg)
 
     @log_step("Get container push items")
     def get_docker_push_items(self):
@@ -147,8 +136,15 @@ class PushDocker:
                 url_items[pull_url].append(item)
             else:
                 url_items[pull_url] = [item]
-        for _, items in sorted(url_items.items()):
-            docker_push_items.append(items[0])
+        for _, items in url_items.items():
+            non_amd64 = True
+            for item in items:
+                if item.metadata["arch"] in ["amd64", "x86_64"]:
+                    docker_push_items.append(item)
+                    non_amd64 = False
+                    break
+            if non_amd64:
+                docker_push_items.append(items[0])
 
         return docker_push_items
 
@@ -255,20 +251,22 @@ class PushDocker:
             )
 
         for repo in repos:
-            LOG.info("Checking validity of Comet repository '{0}'".format(repo))
-            # Check if repo exists in Comet
-            try:
-                metadata = cls.get_repo_metadata(repo, target_settings)
-            except requests.exceptions.HTTPError as e:
-                if e.response.status_code == 404:
-                    raise InvalidRepository("Repository {0} doesn't exist in Comet".format(repo))
-                else:
-                    raise
-
-            # Check if repo is not deprecated
-            # TODO: check with Comet team if this is a reliable way of checking
-            if "Deprecated" in metadata["release_categories"]:
-                raise InvalidRepository("Repository {0} is deprecated".format(repo))
+            # Only check Pyxis if the option is enabled in target settings
+            if target_settings.get("do_repo_deprecation_check", False):
+                LOG.info("Checking validity of repository metadata '{0}'".format(repo))
+                # Check if repo exists in Pyxis
+                try:
+                    metadata = cls.get_repo_metadata(repo, target_settings)
+                    # Check if repo is not deprecated
+                    if "Deprecated" in metadata.get("release_categories", []):
+                        raise InvalidRepository("Repository {0} is deprecated".format(repo))
+                except requests.exceptions.HTTPError as e:
+                    if e.response.status_code == 404:
+                        LOG.warning(
+                            "Metadata of repository '{0}' don't exist in Pyxis".format(repo)
+                        )
+                    else:
+                        raise
 
             # if we're pushing to prod target, check if repo exists on stage as well
             if "propagated_from" in target_settings:
@@ -331,16 +329,37 @@ class PushDocker:
                     # repo doesn't exist, add to rollback tags
                     if not repo_tags:
                         # for rollback tags digest is not known
-                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None))
+                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None, None))
                         continue
                     # tag exists in the repo, add to backup tags
                     if tag in repo_tags.get("tags", {}):
                         image_tag = image_schema_tag.format(
                             host=self.quay_host, repo=full_repo, tag=tag
                         )
-                        digest = self.dest_quay_client.get_manifest_digest(image_tag)
+                        # it's possible that there will be a mismatch between the two calls
+                        try:
+                            digest = self.dest_quay_client.get_manifest_digest(image_tag)
+                        except requests.exceptions.HTTPError as e:
+                            # When robot account is used, 401 may be returned instead of 404
+                            if e.response.status_code == 404 or e.response.status_code == 401:
+                                digest = self._poll_tag_inconsistency(full_repo, tag)
+                            else:
+                                raise
+
+                        # It was determined that the image doesn't actually exist
+                        if not digest:
+                            rollback_tags.append(PushDocker.ImageData(full_repo, tag, None, None))
+                            continue
+
+                        # skip getting v2s1 for non-amd64 image
+                        if item.metadata["arch"] in ["amd64", "x86_64"]:
+                            v2s1_digest = self.dest_quay_client.get_manifest_digest(
+                                image_tag, media_type=QuayClient.MANIFEST_V2S1_TYPE
+                            )
+                        else:
+                            v2s1_digest = None
                         # for backup tags store also digest
-                        image_data = PushDocker.ImageData(full_repo, tag, digest)
+                        image_data = PushDocker.ImageData(full_repo, tag, digest, v2s1_digest)
                         image = image_schema.format(
                             host=self.quay_host,
                             repo=full_repo,
@@ -351,12 +370,62 @@ class PushDocker:
                     # tag doesn't exist in the repo, add to rollback tags
                     else:
                         # for rollback tags digest is not known
-                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None))
+                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None, None))
 
         # it's possible that rollback tags will contain duplicate entries
         rollback_tags = sorted(list(set(rollback_tags)))
         return (backup_tags, rollback_tags)
 
+    def _poll_tag_inconsistency(self, full_repo, tag, poll_rate=30, timeout=120):
+        """
+        Resolve an inconsistency between repo tags reported by Docker API and tag actually existing.
+
+        This scenario occurs when Docker API claims that a tag exists, but attempting to get the
+        tag's digest results in a 404. It could be caused by a delay between two data sources
+        becoming consistent, or by repo tag being removed between the two calls. Attempt to
+        reconcile the difference by periodically calling the two sources. If the data is still
+        inconsistent after timeout, assume that the image doesn't exist and continue.
+
+        Args:
+            full_repo (str):
+                Repo name to poll for tags
+            tag (str):
+                Tag to poll for manifest digest.
+            poll_rate (int, optional):
+                How many seconds to wait between polling. Defaults to 30.
+            timeout (int, optional):
+                How many seconds to poll before giving up and assuming that the image doesn't exist.
+                Defaults to 120.
+
+        Returns (str, None):
+            Image digest, or None if digest doesn't exist (or timeout is reached).
+        """
+        image_schema_tag = "{host}/{repo}:{tag}"
+        image_tag = image_schema_tag.format(host=self.quay_host, repo=full_repo, tag=tag)
+
+        for i in range(timeout // poll_rate):
+            sleep(poll_rate)
+            repo_tags = self.dest_quay_client.get_repository_tags(full_repo)
+
+            try:
+                digest = self.dest_quay_client.get_manifest_digest(image_tag)
+            except requests.exceptions.HTTPError as e:
+                # When robot account is used, 401 may be returned instead of 404
+                if e.response.status_code == 404 or e.response.status_code == 401:
+                    digest = None
+                else:
+                    raise
+
+            if tag in repo_tags.get("tags", {}) and digest:
+                return digest
+            if tag not in repo_tags.get("tags", {}) and digest is None:
+                return None
+
+        LOG.warning(
+            "Unable to determine if image '{0}' exists, assuming it doesn't.".format(image_tag)
+        )
+        return None
+
     @log_step("Perform rollback")
     def rollback(self, backup_tags, rollback_tags):
         """
@@ -392,13 +461,14 @@ class PushDocker:
     def remove_old_signatures(
         self,
         push_items,
-        operator_push_items,
         existing_index_images,
         iib_results,
         backup_tags,
         container_signature_handler,
         operator_signature_handler,
         signature_remover,
+        claim_messages,
+        ii_claim_messages,
     ):
         """
         Remove signatures of containers for tags which were overwritten in the current push.
@@ -413,8 +483,6 @@ class PushDocker:
         Args:
             push_items ([_PushItem]):
                 List of container push items.
-            operator_push_items ([_PushItem]):
-                List of operator push items.
             existing_index_images: ([(digest, version)]):
                 List of tuple of digest + version(tag) of index images which existed
                 before new index image was pushed in the current task
@@ -429,10 +497,12 @@ class PushDocker:
                 ContanerSignatureHandler instance.
             operator_signature_handler (SignatureRemover):
                 SignatureRemover instance.
+            claims_messages (list(dict)):
+                claim messages created for new container items
+            ii_claims_message (list(dict)):
+                claim messages created for new index image(s)
+
         """
-        claim_messages = []
-        for item in push_items:
-            claim_messages += container_signature_handler.construct_item_claim_messages(item)
         new_signatures = [(m["manifest_digest"], m["docker_reference"]) for m in claim_messages]
         outdated_signatures = []
 
@@ -443,20 +513,23 @@ class PushDocker:
                     outdated_signatures.append((arch_manifest["digest"], image_data.tag, ext_repo))
             else:
                 outdated_signatures.append((image_data.digest, image_data.tag, ext_repo))
+            # also add V2S1 signature (only one per tag)
+            if image_data.v2s1_digest:
+                outdated_signatures.append((image_data.v2s1_digest, image_data.tag, ext_repo))
 
         signatures_to_remove = []
-        for esig in container_signature_handler.get_signatures_from_pyxis(
+        for existing_signature in container_signature_handler.get_signatures_from_pyxis(
             [sig[0] for sig in outdated_signatures]
         ):
             if (
-                esig["manifest_digest"],
-                esig["reference"].split(":")[-1],
-                esig["repository"],
+                existing_signature["manifest_digest"],
+                existing_signature["reference"].split(":")[-1],
+                existing_signature["repository"],
             ) in outdated_signatures and (
-                esig["manifest_digest"],
-                esig["reference"],
+                existing_signature["manifest_digest"],
+                existing_signature["reference"],
             ) not in new_signatures:
-                signatures_to_remove.append(esig["_id"])
+                signatures_to_remove.append(existing_signature["_id"])
 
         cert, key = get_pyxis_ssl_paths(self.target_settings)
 
@@ -469,38 +542,23 @@ class PushDocker:
             )
 
         signatures_to_remove = []
-        ii_claim_messages = []
         if existing_index_images:
-            for version, iib_details in sorted(iib_results.items()):
-                iib_result = iib_details["iib_result"]
-                signing_keys = iib_details["signing_keys"]
-                image_schema = "{host}/{namespace}/{repo}@{digest}"
-                iib_namespace = iib_result.index_image_resolved.split("/")[1]
-                image_digest = iib_result.index_image_resolved.split("@")[1]
-                intermediate_index_image = image_schema.format(
-                    host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
-                    namespace=iib_namespace,
-                    repo="iib",
-                    digest=image_digest,
-                )
-                ii_claim_messages += (
-                    operator_signature_handler.construct_index_image_claim_messages(
-                        intermediate_index_image, [version], signing_keys
-                    )
-                )
             new_operator_signatures = [
                 (m["manifest_digest"], m["docker_reference"]) for m in ii_claim_messages
             ]
 
-            for esig in container_signature_handler.get_signatures_from_pyxis(
+            for existing_signature in container_signature_handler.get_signatures_from_pyxis(
                 [digest_version[0] for digest_version in existing_index_images]
             ):
-                if (esig["manifest_digest"], esig["reference"]) not in new_operator_signatures and (
-                    esig["manifest_digest"],
-                    esig["reference"].split(":")[-1],
-                    esig["repository"],
+                if (
+                    existing_signature["manifest_digest"],
+                    existing_signature["reference"],
+                ) not in new_operator_signatures and (
+                    existing_signature["manifest_digest"],
+                    existing_signature["reference"].split(":")[-1],
+                    existing_signature["repository"],
                 ) in existing_index_images:
-                    signatures_to_remove.append(esig["_id"])
+                    signatures_to_remove.append(existing_signature["_id"])
             if signatures_to_remove:
                 signature_remover.remove_signatures_from_pyxis(
                     signatures_to_remove,
@@ -509,6 +567,68 @@ class PushDocker:
                     key,
                 )
 
+    def _fetch_digest(self, repo, tag, media_type):
+        """Fetch digest of repo and tag for given media type.
+
+        Args:
+            repo(str):
+                Repository name.
+            tag(str):
+                Image tag.
+            mtype(str):
+                Media type for requested digest.
+                (can be only application/vnd.docker.distribution.manifest.v2+json,
+                application/vnd.docker.distribution.manifest.v1+json)
+
+        Returns(str): Manifest digest of the container.
+        """
+        image_schema = "{host}/{namespace}/{repo}:{tag}"
+        dest_ref = image_schema.format(
+            host=self.quay_host,
+            namespace=self.target_settings["quay_namespace"],
+            repo=repo,
+            tag=tag,
+        )
+        digest = self.dest_quay_client.get_manifest_digest(dest_ref, media_type=media_type)
+        return digest
+
+    def fetch_missing_push_items_digests(self, push_items, target_settings):
+        """Fetch digests for media types which weren't originally pushed.
+
+        In order to be able to sign v2s1 for images which were pushed as
+        v2s2 or to sign v2s2 for images which were pushed as v2s1
+        fetch digests of those missing media types from quay and
+        set it to item metadata into 'new_digests' mapping.
+
+        Args:
+            push_items(list): List of push items.
+            target_settings(dict): Target settings.
+        """
+        for item in push_items:
+            item.metadata["new_digests"] = {}
+            missing_media_types = set(
+                [QuayClient.MANIFEST_V2S2_TYPE, QuayClient.MANIFEST_V2S1_TYPE]
+            ) - set(item.metadata["build"]["extra"]["image"]["media_types"])
+
+            # Always add v2s1 due to possible digest change
+            missing_media_types.add(QuayClient.MANIFEST_V2S1_TYPE)
+            for repo, tags in item.metadata["tags"].items():
+                internal_repo = get_internal_container_repo_name(repo)
+                v2_sch2_cache = {}
+                for tag in tags:
+                    for mtype in missing_media_types:
+                        if mtype == QuayClient.MANIFEST_V2S2_TYPE:
+                            if repo not in v2_sch2_cache:
+                                v2_sch2_cache[repo] = self._fetch_digest(internal_repo, tag, mtype)
+                            item.metadata["new_digests"].setdefault((repo, tag), {})[
+                                mtype
+                            ] = v2_sch2_cache[repo]
+                        # Fetch v2s1 only for amd64 image
+                        elif item.metadata["arch"] in ["amd64", "x86_64"]:
+                            item.metadata["new_digests"].setdefault((repo, tag), {})[
+                                mtype
+                            ] = self._fetch_digest(internal_repo, tag, mtype)
+
     def run(self):
         """
         Perform the full push-docker workflow.
@@ -519,6 +639,8 @@ class PushDocker:
         - Generate backup mapping that will be used for rollback if something goes wrong.
         - Sign container images using RADAS and upload signatures to Pyxis
         - Push container images to their destinations
+        - Fetching digests for missing media types of pushed items
+        - Sign manifests for missing media types (has to be done after pushing)
         - Filter out push items to only include operator image items
         - Add operator bundles to index images by using IIB
         - Sign index images using RADAS and upload signatures to Pyxis
@@ -537,50 +659,74 @@ class PushDocker:
         backup_tags, rollback_tags = self.generate_backup_mapping(docker_push_items)
         existing_index_images = []
         iib_results = None
+        successful_iib_results = dict()
         index_stamp = timestamp()
-        try:
-            # Sign container images
-            container_signature_handler = ContainerSignatureHandler(
-                self.hub, self.task_id, self.target_settings, self.target_name
-            )
-            operator_signature_handler = OperatorSignatureHandler(
-                self.hub, self.task_id, self.target_settings, self.target_name
-            )
-            sig_remover = SignatureRemover()
-            sig_remover.set_quay_client(self.dest_quay_client)
+        # Sign container images
+        container_signature_handler = ContainerSignatureHandler(
+            self.hub, self.task_id, self.target_settings, self.target_name
+        )
+        operator_signature_handler = OperatorSignatureHandler(
+            self.hub, self.task_id, self.target_settings, self.target_name
+        )
+        sig_remover = SignatureRemover()
+        sig_remover.set_quay_client(self.dest_quay_client)
 
-            container_signature_handler.sign_container_images(docker_push_items)
-            # Push container images
-            container_pusher = ContainerImagePusher(docker_push_items, self.target_settings)
-            container_pusher.push_container_images()
+        manifest_claims = container_signature_handler.sign_container_images(docker_push_items)
+        # Push container images
+        container_pusher = ContainerImagePusher(docker_push_items, self.target_settings)
+        container_pusher.push_container_images()
 
-            if operator_push_items:
-                # Build index images
-                operator_pusher = OperatorPusher(operator_push_items, self.target_settings)
-                existing_index_images = operator_pusher.get_existing_index_images(
-                    self.dest_quay_client
-                )
-                iib_results = operator_pusher.build_index_images()
-                # Sign operator images
-                operator_signature_handler.sign_operator_images(iib_results, index_stamp)
-                # Push index images to Quay
-                operator_pusher.push_index_images(iib_results, index_stamp)
-        except (Exception, SystemExit):
-            LOG.error("An exception has occurred during the push, starting rollback")
-            self.rollback(backup_tags, rollback_tags)
-            raise
-        else:
-            # Remove old signatures
-            self.remove_old_signatures(
-                docker_push_items,
-                operator_push_items,
-                existing_index_images,
-                iib_results,
-                backup_tags,
-                container_signature_handler,
-                operator_signature_handler,
-                sig_remover,
+        # fetch missing digests
+        self.fetch_missing_push_items_digests(docker_push_items, self.target_settings)
+
+        # sign missing images
+        manifest_claims += container_signature_handler.sign_container_images_new_digests(
+            docker_push_items
+        )
+
+        failed = False
+
+        ii_manifest_claims = []
+        if operator_push_items:
+            # Build index images
+            operator_pusher = OperatorPusher(
+                operator_push_items, self.task_id, self.target_settings
             )
+            existing_index_images = operator_pusher.get_existing_index_images(self.dest_quay_client)
+            iib_results = operator_pusher.build_index_images()
+            # Sign operator images
+            successful_iib_results = dict(
+                [(key, val) for key, val in iib_results.items() if val["iib_result"]]
+            )
+            ii_manifest_claims += operator_signature_handler.sign_operator_images(
+                successful_iib_results, index_stamp
+            )
+            # Push index images to Quay
+            operator_pusher.push_index_images(successful_iib_results, index_stamp)
+            # Rollback only when all index image builds fails
+            if not any([x["iib_result"] for x in iib_results.values()]):
+                LOG.error("Push of all index images failed, running rollback.")
+                self.rollback(backup_tags, rollback_tags)
+                failed = True
+            if successful_iib_results != iib_results:
+                LOG.error("Push of some index images failed")
+                failed = True
+
+        # Remove old signatures
+        self.remove_old_signatures(
+            docker_push_items,
+            existing_index_images,
+            dict([(k, v) for k, v in successful_iib_results.items() if v["iib_result"]]),
+            backup_tags,
+            container_signature_handler,
+            operator_signature_handler,
+            sig_remover,
+            manifest_claims,
+            ii_manifest_claims,
+        )
+        if failed:
+            # Why???
+            sys.exit(1)
 
 
 def mod_entry_point(push_items, hub, task_id, target_name, target_settings):
