@@ -1,16 +1,17 @@
 import base64
-from datetime import datetime
 import json
 import logging
 import uuid
 import tempfile
 
-import proton
-
 from .exceptions import SigningError
-from .utils.misc import run_entrypoint, log_step, get_pyxis_ssl_paths
+from .utils.misc import (
+    run_entrypoint,
+    log_step,
+    get_pyxis_ssl_paths,
+)
 from .quay_client import QuayClient
-from .manifest_claims_handler import ManifestClaimsHandler
+from .manifest_claims_handler import _ManifestClaimsRunner, UMBSettings
 
 LOG = logging.getLogger("pubtools.quay")
 
@@ -49,6 +50,7 @@ class SignatureHandler:
 
         self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
         self._src_quay_client = None
+        self._dest_quay_client = None
 
     @property
     def src_quay_client(self):
@@ -57,10 +59,21 @@ class SignatureHandler:
             self._src_quay_client = QuayClient(
                 self.target_settings["source_quay_user"],
                 self.target_settings["source_quay_password"],
-                self.quay_host,
+                self.target_settings.get("source_quay_host") or self.quay_host,
             )
         return self._src_quay_client
 
+    @property
+    def dest_quay_client(self):
+        """Create and access QuayClient for dest image."""
+        if self._dest_quay_client is None:
+            self._dest_quay_client = QuayClient(
+                self.target_settings["dest_quay_user"],
+                self.target_settings["dest_quay_password"],
+                self.quay_host,
+            )
+        return self._dest_quay_client
+
     @classmethod
     def create_manifest_claim_message(
         cls, destination_repo, signature_key, manifest_digest, docker_reference, image_name, task_id
@@ -109,11 +122,10 @@ class SignatureHandler:
             "repo": destination_repo,
             "image_name": image_name,
             "docker_reference": docker_reference,
-            "created": datetime.utcnow().isoformat() + "Z",
         }
         return message
 
-    def get_tagged_image_digests(self, image_ref):
+    def get_tagged_image_digests(self, image_ref, media_type):
         """
         Get all digests referenced by a tagged image.
 
@@ -123,16 +135,19 @@ class SignatureHandler:
         Args:
             image_ref (str):
                 Image reference URL. Must be specified via tag.
+            media_type (str):
+                Requested media type of the manifest.
 
         Returns ([str]):
             List of manifest digests referenced by the tag.
         """
         digests = []
-
-        manifest = self.src_quay_client.get_manifest(image_ref)
+        manifest = self.src_quay_client.get_manifest(image_ref, media_type=media_type)
         # If V2S2 manifest, we only want its digest
-        if manifest["mediaType"] == "application/vnd.docker.distribution.manifest.v2+json":
-            digests.append(self.src_quay_client.get_manifest_digest(image_ref))
+        if manifest.get("mediaType") != "application/vnd.docker.distribution.manifest.list.v2+json":
+            digests.append(
+                self.src_quay_client.get_manifest_digest(image_ref, media_type=media_type)
+            )
         # If manifest list, we want digests of all its arch images
         else:
             for arch_manifest in manifest["manifests"]:
@@ -286,25 +301,24 @@ class SignatureHandler:
 
         address = (
             "queue://Consumer.msg-producer-pub"
-            ".{task_id}.VirtualTopic.eng.robosignatory.container.sign".format(task_id=self.task_id)
+            ".{task_id}.VirtualTopic.eng.robosignatory.container.sign.{task_id}".format(
+                task_id=self.task_id
+            )
         )
-
         docker_settings = self.target_settings["docker_settings"]
-        claims_handler = ManifestClaimsHandler(
-            umb_urls=docker_settings["umb_urls"],
+        umb_settings = UMBSettings(
+            broker_urls=docker_settings["umb_urls"],
             radas_address=docker_settings.get("umb_radas_address", address),
-            claim_messages=claim_messages,
             pub_cert=docker_settings.get("umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"),
             ca_cert=docker_settings.get("umb_ca_cert", "/etc/pki/tls/certs/ca-bundle.crt"),
-            timeout=docker_settings.get("umb_signing_timeout", 600),
-            throttle=docker_settings.get("umb_signing_throttle", 100),
-            retry=docker_settings.get("umb_signing_retry", 3),
-            message_sender_callback=message_sender_callback,
+            signing_timeout=docker_settings.get("umb_signing_timeout", 600),
+            signing_throttle=docker_settings.get("umb_signing_throttle", 100),
+            signing_retry=docker_settings.get("umb_signing_retry", 3),
         )
-        container = proton.reactor.Container(claims_handler)
-        container.run()
 
-        return claims_handler.received_messages
+        runner = _ManifestClaimsRunner(umb_settings, claim_messages, message_sender_callback)
+        runner.start()
+        return runner.received_messages
 
     def upload_signatures_to_pyxis(self, claim_mesages, signature_messages):
         """
@@ -415,7 +429,19 @@ class ContainerSignatureHandler(SignatureHandler):
         claim_messages = []
 
         if push_item.claims_signing_key:
-            digests = self.get_tagged_image_digests(push_item.metadata["pull_url"])
+            if (
+                QuayClient.MANIFEST_LIST_TYPE
+                in push_item.metadata["build"]["extra"]["image"]["media_types"]
+            ):
+                mtype = QuayClient.MANIFEST_LIST_TYPE
+            elif (
+                QuayClient.MANIFEST_V2S2_TYPE
+                in push_item.metadata["build"]["extra"]["image"]["media_types"]
+            ):
+                mtype = QuayClient.MANIFEST_V2S2_TYPE
+            else:
+                mtype = QuayClient.MANIFEST_V2S1_TYPE
+            digests = self.get_tagged_image_digests(push_item.metadata["pull_url"], mtype)
             # each image digest needs its own signature
             for digest in digests:
                 # each destination image reference needs its own signature
@@ -479,12 +505,14 @@ class ContainerSignatureHandler(SignatureHandler):
         Args:
             push_items (([ContainerPushItem])):
                 Container push items whose images will be signed.
+        Returns (list(dict)):
+            List of claim message sent for signing
         """
         if not self.target_settings["docker_settings"].get(
             "docker_container_signing_enabled", False
         ):
             LOG.info("Container signing not allowed in target settings, skipping.")
-            return
+            return []
 
         claim_messages = []
         for item in push_items:
@@ -493,12 +521,53 @@ class ContainerSignatureHandler(SignatureHandler):
         claim_messages = self.filter_claim_messages(claim_messages)
         if not claim_messages:
             LOG.info("No new claim messages will be uploaded")
-            return
+            return []
 
         LOG.info("{0} claim messages will be uploaded".format(len(claim_messages)))
         signature_messages = self.get_signatures_from_radas(claim_messages)
         self.validate_radas_messages(claim_messages, signature_messages)
         self.upload_signatures_to_pyxis(claim_messages, signature_messages)
+        return claim_messages
+
+    @log_step("Sign container images")
+    def sign_container_images_new_digests(self, push_items):
+        """
+        Sign digests for manifest types which haven't existed before the push.
+
+        Manifest types which are not pushed in the push items can be fetched
+        from quay and then created dynamically on the fly. This method is meant
+        to create signatures for those.
+
+        Args:
+            push_items (([ContainerPushItem])):
+                Container push items whose images will be signed.
+        Returns (list(dict)):
+            List of claim message sent for signing
+        """
+        if not self.target_settings["docker_settings"].get(
+            "docker_container_signing_enabled", False
+        ):
+            LOG.info("Container signing not allowed in target settings, skipping.")
+            return []
+
+        claim_messages = []
+        for item in push_items:
+            for (repo, tag), digests_per_type in sorted(item.metadata["new_digests"].items()):
+                for digest in digests_per_type.values():
+                    claim_messages += self.construct_variant_claim_messages(
+                        repo, tag, digest, [item.claims_signing_key]
+                    )
+        claim_messages = self.remove_duplicate_claim_messages(claim_messages)
+        claim_messages = self.filter_claim_messages(claim_messages)
+        if not claim_messages:
+            LOG.info("No new claim messages will be uploaded")
+            return []
+
+        LOG.info("{0} claim messages will be uploaded".format(len(claim_messages)))
+        signature_messages = self.get_signatures_from_radas(claim_messages)
+        self.validate_radas_messages(claim_messages, signature_messages)
+        self.upload_signatures_to_pyxis(claim_messages, signature_messages)
+        return claim_messages
 
 
 class OperatorSignatureHandler(SignatureHandler):
@@ -529,7 +598,9 @@ class OperatorSignatureHandler(SignatureHandler):
             index_image_credential[1],
             self.quay_host,
         )
-        manifest_list = index_image_quay_client.get_manifest(index_image, manifest_list=True)
+        manifest_list = index_image_quay_client.get_manifest(
+            index_image, media_type=QuayClient.MANIFEST_LIST_TYPE
+        )
         digests = [m["digest"] for m in manifest_list["manifests"]]
         for registry in self.dest_registries:
             for signing_key in signing_keys:
@@ -563,34 +634,33 @@ class OperatorSignatureHandler(SignatureHandler):
             iib_results ({str:dict}):
                 IIB results for each version the push was performed for.
         """
-        image_schema = "{host}/{namespace}/{repo}@{digest}"
+        image_schema = "{host}/{namespace}/{repo}:{tag}"
         if not self.target_settings["docker_settings"].get(
             "docker_container_signing_enabled", False
         ):
             LOG.info("Container signing not allowed in target settings, skipping.")
-            return
+            return []
 
         claim_messages = []
         for version, iib_details in sorted(iib_results.items()):
             iib_result = iib_details["iib_result"]
             signing_keys = iib_details["signing_keys"]
-            # Using intermediate index image to ensure that it doesn't get overwritten
+            # Index image used to fetch manifest list. This image will never be overwritten
             iib_namespace = iib_result.index_image_resolved.split("/")[1]
-            image_digest = iib_result.index_image_resolved.split("@")[1]
-            intermediate_index_image = image_schema.format(
+            permanent_index_image = image_schema.format(
                 host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
                 namespace=iib_namespace,
                 repo="iib",
-                digest=image_digest,
+                tag=iib_result.build_tags[0],
             )
             # Version acts as a tag of the index image
             claim_messages += self.construct_index_image_claim_messages(
-                intermediate_index_image, [version, "%s-%s" % (version, tag_suffix)], signing_keys
+                permanent_index_image, [version, "%s-%s" % (version, tag_suffix)], signing_keys
             )
 
         if not claim_messages:
             LOG.info("No new claim messages will be uploaded")
-            return
+            return []
 
         signature_messages = self.get_signatures_from_radas(claim_messages)
         self.validate_radas_messages(claim_messages, signature_messages)
