@@ -584,7 +584,9 @@ class TagDocker:
 
         signature_handler.sign_claim_messages(claim_messages, True, True)
 
-        raw_src_manifest = self.quay_client.get_manifest(source_image, manifest_list=True, raw=True)
+        raw_src_manifest = self.quay_client.get_manifest(
+            source_image, media_type=QuayClient.MANIFEST_LIST_TYPE, raw=True
+        )
 
         # Special case: if the source manifest and the merged manifest are the same, upload the
         # raw source manifest. The reason is that otherwise the digests of the copied manifests
@@ -617,7 +619,6 @@ class TagDocker:
             remove_last=remove_last,
             quay_user=target_settings["dest_quay_user"],
             quay_password=target_settings["dest_quay_password"],
-            send_umb_msg=False,
         )
 
     def untag_image(self, push_item, tag):
@@ -672,7 +673,9 @@ class TagDocker:
             host=self.quay_host, namespace=namespace, repo=internal_repo
         )
         dest_image = "{0}:{1}".format(full_repo, tag)
-        manifest_list = self.quay_client.get_manifest(dest_image, manifest_list=True)
+        manifest_list = self.quay_client.get_manifest(
+            dest_image, media_type=QuayClient.MANIFEST_LIST_TYPE
+        )
         keep_manifests = []
 
         for manifest in manifest_list["manifests"]:
@@ -725,9 +728,13 @@ class TagDocker:
             self.target_settings.get("docker_timeout"),
             self.target_settings.get("docker_tls_verify") or False,
             self.target_settings.get("docker_cert_path") or None,
+            self.target_settings.get("skopeo_executor_username") or None,
+            self.target_settings.get("skopeo_executor_password") or None,
         ) as executor:
             executor.skopeo_login(
-                self.target_settings["dest_quay_user"], self.target_settings["dest_quay_password"]
+                self.quay_host,
+                self.target_settings["dest_quay_user"],
+                self.target_settings["dest_quay_password"],
             )
             for item in self.push_items:
                 for tag in item.metadata["add_tags"]:
